#include <stdio.h>
#include <string.h> /* strdup() */
#include <stdlib.h> /* exit() */

#include "node.h"

int print_dummy=0;
int debug_messages=0;

struct node_list {
   char *name;
   struct node *node;
   struct node_list *next;
   int printed;
};

static struct node_list head_node_list={NULL,NULL,NULL};

   /* Frees an unused node */
void free_node(struct node *v) {
   if (v==NULL) {
      printf("Freeing NULL node!\n");
      return;
   }
   	
   free(v->name);
   free(v);
}



/* find the node that is the representative of the set that v is in */
struct node *rep_(struct node *v) {

   struct node *temp_node;

//   if (debug_messages) {
//      fprintf(stderr,"Entering rep\n");
//   }
   
   if (v==NULL) {
      return NULL;
   }
      
   temp_node=v;
   while(temp_node->rep!=NULL) {
      temp_node=temp_node->rep;
   }

//   if (debug_messages) {
//      fprintf(stderr,"Leaving rep\n");
//   }
   
   return temp_node;   
}

   /* Prints the name of the node, or "bottom" if NULL */
void dump_node_name(FILE *output,struct node *v) {
   if (v==NULL) fprintf(output,"bottom");
   else fprintf(output,"%s",v->name);
}

   
   /* Prints the names of nodes pointed to */
static void dump_points_to(FILE *output,struct node *v) {
   
   struct node *temp_node;
   
   temp_node=v->points_to;
   
   if (temp_node==NULL) {
      fprintf(output,"bottom");
   }
   else if ( (rep_(temp_node)->dummy) && (!print_dummy)) {
      fprintf(output,"bottom");
   }
   else {
      temp_node=rep_(temp_node);
      fprintf(output,"%s",temp_node->name);
   }
   
}

#define DOT "/usr/bin/dot"

   /* Makes a graph from the output.  Uses "dot" from */
   /* www.graphvz.org                                 */
void dump_nodes_html(char *html_prefix,int line_count) {
   
   struct node_list *temp_list,*temp_list2;
   struct node *rep,*temp_node;
   FILE *dot_pipe;
   
   char temp_string[BUFSIZ];
   
   sprintf(temp_string,"%s -Tpng -o%s_%i.png",DOT,html_prefix,line_count);   
   dot_pipe=popen(temp_string,"w");
   //sprintf(temp_string,"%s_%i.dot",html_prefix,line_count);
   //dot_pipe=fopen(temp_string,"w");
   if (dot_pipe==NULL) {
      fprintf(stderr,"Warning!  Could not run DOT for making graphs %s\n",
	      temp_string);
      return;
   }
   
   fprintf(dot_pipe,"digraph G {\n");
      
   temp_list=head_node_list.next;
   while(temp_list!=NULL) {
      temp_list->printed=0;
      temp_list=temp_list->next;
   }
   
   temp_list=head_node_list.next;
   while(temp_list!=NULL) {
      if ( (!temp_list->printed) && (!temp_list->node->dummy || print_dummy)) {
	 
	 rep=rep_(temp_list->node);

	 fprintf(dot_pipe,"\t%s [label=\"",rep->name);
	 
	 temp_list2=temp_list;
	 while(temp_list2!=NULL) {
	    if (rep_(temp_list2->node)==rep) {
	       if (!temp_list2->node->dummy || print_dummy) {
		  fprintf(dot_pipe,"%s",temp_list2->node->name);
	       }
	       fprintf(dot_pipe," ");
	       temp_list2->printed=1;
	    }
	    temp_list2=temp_list2->next;
	 }
	 fprintf(dot_pipe,"\"];\n");
	 
	 if (rep->points_to!=NULL) {
	    if (! rep_(rep->points_to)->dummy || print_dummy) {
	       temp_node=rep_(rep->points_to);
	       fprintf(dot_pipe,"\t%s -> ",rep->name);
	       dump_points_to(dot_pipe,rep);
	       fprintf(dot_pipe,"\n");
	    }
	 }
      }
      
      temp_list=temp_list->next;
   }
   fprintf(dot_pipe,"}\n");
   fclose(dot_pipe);
}

   /* Prints out all the nodes and their relations */
void dump_nodes(FILE *output) {
   struct node_list *temp_list,*temp_list2;
   struct node *rep;
   
   fprintf(output,"=Dump begin================\n");
   
     /* Walk through list, making sure to only print nodes */
     /* we haven't seen before                             */
   temp_list=head_node_list.next;
   while(temp_list!=NULL) {
      temp_list->printed=0;
      temp_list=temp_list->next;
   }
   
   
   temp_list=head_node_list.next;
   while(temp_list!=NULL) {
      if ( (!temp_list->printed) && (!temp_list->node->dummy || print_dummy)) {
	 rep=rep_(temp_list->node);

	 fprintf(output,"( ");
	 
	 /* print the rep of the node surrounded by * */
	 /* print the rest in a line */
	 /* only print dummy variables if specified */
	 temp_list2=temp_list;
	 while(temp_list2!=NULL) {
	    if (temp_list2->node==rep) fprintf(output,"*");
	    if (rep_(temp_list2->node)==rep) {
	       if (!temp_list2->node->dummy || print_dummy) {
		  fprintf(output,"%s",temp_list2->node->name);
	       }
	       if (temp_list2->node==rep) fprintf(output,"*");
	       fprintf(output," ");
	       temp_list2->printed=1;
	    }
	    temp_list2=temp_list2->next;
	 }
	 fprintf(output,")\t");
	 
	    /* print the points to for this node */
	 fprintf(output,"Points_to=");
	 dump_points_to(output,rep);
	 fprintf(output,"\n");
	 
      }
      
      temp_list=temp_list->next;
   }
   printf("=Dump end================\n");   
   
}

   
   /* create a node */
struct node *new_node(char *name) {

   struct node *temp_node;
   struct node_list *temp_list;
   
   temp_node=calloc(1,sizeof(struct node));
   if (temp_node==NULL) {
      printf("Error allocating space for node!\n");
      exit(1);
   }
   
   temp_node->name=strdup(name);
   temp_node->dummy=0;
   
   temp_list=&head_node_list;
   while(temp_list->next!=NULL) temp_list=temp_list->next;
   temp_list->next=calloc(1,sizeof(struct node_list));
   if (temp_list->next==NULL) {
      printf("Error allocating space for node_list!\n");
      exit(1);
   }
   temp_list=temp_list->next;
   temp_list->name=strdup(name);
   temp_list->node=temp_node;
   temp_list->next=NULL;

   return temp_node;
}

  /* make a dummy node */
struct node *make_dummy() {
   
      char temp_string[BUFSIZ];
      static int dummies=0;
      struct node *temp_node;
   
      sprintf(temp_string,"dummy%i",dummies);
      dummies++;
      temp_node=new_node(temp_string);
      temp_node->dummy=1;
   
      return temp_node;
}


  /* given a string, find the corresponding node */
  /* linear search, could be made faster */
struct node *find_node(char *name) {
   
   struct node_list *temp_list;
   
   temp_list=head_node_list.next;

   while(temp_list!=NULL) {
      if (!strcmp(name,temp_list->name)) {
	 return temp_list->node;
      }
      temp_list=temp_list->next;
   }
   
   return NULL;
   
}

   /* crate the union of two nodes, returning the rep     */
   /* name has an underscore due to C namespace collision */
struct node *union_(struct node *v1, struct node *v2) {

   if (debug_messages) {
      fprintf(stderr,"Union: ");
      dump_node_name(stderr,v1);
      fprintf(stderr,"+");
      dump_node_name(stderr,v2);
      fprintf(stderr,"\n");
   }
   
   if ((v2==NULL) && (v1==NULL)) return NULL;
   if (v2==NULL) return v1;
   if (v1==NULL) return v2;
   if (v1==v2) return rep_(v1);
   
   if (v1->dummy) {
      v1->rep=rep_(v2);
   }
   else {
      v2->rep=rep_(v1);
   }
   
   if (debug_messages) {
      fprintf(stderr,"Done Union\n");
   }
   
   return rep_(v1);
}

  /* return points-to info for a node */
struct node *pt(struct node *v) {

   struct node *temp_node;
      
   temp_node=rep_(v);
   
   if (temp_node==NULL) {
      return NULL;
   }
   
   return temp_node->points_to;
}

  /* perform a recursive union of two vars */
struct node *rec_union(struct node *v1, struct node *v2) {

   struct node *p1,*p2,*t1,*t2;
   
   if (debug_messages) {
      fprintf(stderr,"rec_union: ");
      dump_node_name(stderr,v1);
      fprintf(stderr," ");
      dump_node_name(stderr,v2);
      fprintf(stderr,"\n");
   }
   
   p1=pt(rep_(v1));
   p2=pt(rep_(v2));
      
   t1=union_(rep_(v1),rep_(v2));
   
   if (debug_messages) {
      fprintf(stderr,"Found union!\n");
   }
   
   if (p1==p2) return t1; /* ? */
   else if (p1!=NULL && p2!=NULL) t2=rec_union(p1,p2);
   else if (p1!=NULL) t2=p1;
   else if (p2!=NULL) t2=p2;
   else t2=NULL;

   t1->points_to=t2;
   
   return t1;
   
}
